#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sched.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <unistd.h>
#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_mbuf.h>
#include <rte_ring.h>
#include <rte_eventdev.h>
#include <rte_dma.h>
#include <infiniband/verbs.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define SERVER "example.com"
#define PORT 443
#define NUM_THREADS 512  
#define NUM_REQUESTS 1000000
#define REQUEST "GET / HTTP/1.1\r\nHost: example.com\r\nConnection: close\r\n\r\n"
#define NIC_PORT 0  
#define BURST_SIZE 524288  
#define NUM_TX_QUEUES NUM_THREADS  
#define EVENT_DEV_ID 0

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
SSL_CTX* ctx;
struct rte_mempool *mbuf_pool;
struct rte_ring *tx_ring;
struct rte_dma_conf dma_conf;
struct ibv_context *rdma_ctx;
struct ibv_pd *rdma_pd;
struct ibv_mr *rdma_mr;
struct ibv_qp *rdma_qp;
struct rte_eventdev_config event_conf;

// **NUMAノードごとのスレッド割り当て**
void bind_thread_to_numa(int thread_id) {
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    int cpu_id = thread_id % sysconf(_SC_NPROCESSORS_ONLN);
    CPU_SET(cpu_id, &cpuset);
    pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpuset);
}

// **RDMAメモリ領域の登録**
void initialize_rdma() {
    struct ibv_device **dev_list = ibv_get_device_list(NULL);
    rdma_ctx = ibv_open_device(dev_list[0]);
    rdma_pd = ibv_alloc_pd(rdma_ctx);

    void *rdma_buf = aligned_alloc(4096, BURST_SIZE);
    rdma_mr = ibv_reg_mr(rdma_pd, rdma_buf, BURST_SIZE, IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
}

// **BoringSSL + TLS 1.3 + 耐量子暗号の設定**
SSL_CTX* initialize_tls() {
    SSL_CTX* ctx = SSL_CTX_new(TLS_client_method());

    SSL_CTX_set_min_proto_version(ctx, TLS1_3_VERSION);
    SSL_CTX_set_cipher_list(ctx, "TLS_AES_256_GCM_SHA384");
    SSL_CTX_set1_groups_list(ctx, "kyber512:kyber768:frodo640aes");

    SSL_CTX_set_mode(ctx, SSL_MODE_ASYNC | SSL_MODE_RELEASE_BUFFERS | SSL_MODE_ENABLE_PARTIAL_WRITE);
    SSL_CTX_set_max_early_data(ctx, 67108864);
    SSL_CTX_set_options(ctx, SSL_OP_NO_TICKET | SSL_OP_NO_QUERY_MTU | SSL_OP_NO_RENEGOTIATION | SSL_OP_NO_CLIENT_CERTS);

    return ctx;
}

// **HTTPSリクエストパケット作成**
void* worker_thread(void* arg) {
    int thread_id = *(int*)arg;
    bind_thread_to_numa(thread_id);

    struct rte_mbuf* pkts[BURST_SIZE];

    for (;;) {
        pthread_mutex_lock(&mutex);
        static int request_count = 0;
        if (request_count >= NUM_REQUESTS) {
            pthread_mutex_unlock(&mutex);
            break;
        }
        request_count += BURST_SIZE;
        pthread_mutex_unlock(&mutex);

        rte_pktmbuf_alloc_bulk(mbuf_pool, pkts, BURST_SIZE);
    }
    return NULL;
}

// **NIC送信 (RDMA)**
void* sender_thread(void* arg) {
    struct ibv_sge sge;
    struct ibv_send_wr wr, *bad_wr;

    while (1) {
        struct rte_mbuf* pkts[BURST_SIZE];
        uint16_t dequeued = rte_event_dequeue_burst(EVENT_DEV_ID, 0, (void **)pkts, BURST_SIZE, NULL);
        if (dequeued == 0) continue;

        sge.addr = (uintptr_t) rdma_mr->addr;
        sge.length = BURST_SIZE;
        sge.lkey = rdma_mr->lkey;

        wr.wr_id = 0;
        wr.sg_list = &sge;
        wr.num_sge = 1;
        wr.opcode = IBV_WR_SEND;
        wr.send_flags = IBV_SEND_INLINE | IBV_SEND_SIGNALED | IBV_SEND_FENCE | IBV_SEND_ACK;

        ibv_post_send(rdma_qp, &wr, &bad_wr);
    }
    return NULL;
}